#!/usr/bin/bash

debug=0

preamble=$(
    if [[ -e diag/front-end.state.awk ]]
    then
	cat diag/front-end.state.awk
    else
	echo "next_slot_loc = 10;"
    fi
)

awk "

BEGIN {
    $preamble
    debug = $debug;
}

"'function slot_alloc(ref_type) {
    new_slot = -1;

    for (new_slot in slot_free) {
        break;
    }

    if (new_slot == -1) {
        new_slot = next_slot_loc;
        ++ next_slot_loc;
    } else {
        delete slot_free[new_slot];
    }

    if (ref_type ~ /temp/) {
        slot_temp[new_slot] = 1;
        slot_status[new_slot] = "temp";
    }

    if (ref_type ~ /var/) {
        slot_var[new_slot] = 1;
        slot_status[new_slot] = "var";
    }

    return new_slot;
}

function dump_state() {
    printf("next_slot_loc=%s\n", next_slot_loc) > "diag/front-end.state.awk";
    for (s in slot_free) printf("slot_free[%s]=1\n", s) >> "diag/front-end.state.awk";
    for (s in slot_temp) printf("slot_temp[%s]=1\n", s) >> "diag/front-end.state.awk";
    for (s in slot_var ) printf("slot_var[%s]=1\n", s) >> "diag/front-end.state.awk";
    for (s in slot_status) printf("slot_status[%s]=\"%s\"\n", s, slot_status[s]) >> "diag/front-end.state.awk";
    for (s in slot_var_name) printf("slot_var_name[%s]=\"%s\"\n", s, slot_var_name[s]) >> "diag/front-end.state.awk";
    for (n in var_slot) printf("var_slot[\"%s\"]=%s\n", n, var_slot[n]) >> "diag/front-end.state.awk";
    for (s in slot_type) printf("slot_type[%s]=\"%s\"\n", s, slot_type[s]) >> "diag/front-end.state.awk";
}

function slot_map() {
    printf("INSTR:print \"next_slot_loc=\",%s,\"\\n\"\n", next_slot_loc);
    for (i=0; i<next_slot_loc; ++i) {
        if (slot_status[i] == "var") {
            printf("INSTR:print \"slot[%s]=%s:%s\\n\"\n",
                   i, slot_status[i], slot_var_name[i]);
        } else {
            printf("INSTR:print \"slot[%s]=%s\\n\"\n", i, slot_status[i]);
        }
    }
}

function slot_gc() {
    for (slot_loc in slot_temp) {
        slot_status[slot_loc] = "free";
        slot_free[slot_loc] = 1;
    }

    delete slot_temp;
}

function expr_split(expression, expr_type) {
    if (debug) printf("INSTR:expr_split(\"%s\",\"%s\")\n", expression, expr_type);

    expr_len = length(expression);

    if (expr_type ~ /^E_PRD$/) repl_pos = match(expression, /@[0-9]+\*@[0-9]+/);
    if (expr_type ~ /^E_DIV$/) repl_pos = match(expression, /@[0-9]+\/@[0-9]+/);
    if (expr_type ~ /^E_SUM$/) repl_pos = match(expression, /@[0-9]+\+@[0-9]+/);
    if (expr_type ~ /^E_DIF$/) repl_pos = match(expression, /@[0-9]+-@[0-9]+/);
    if (expr_type ~ /^E_EXT$/) repl_pos = match(expression, /\[@[0-9]+(,@[0-9]+)*\]/);
    if (expr_type ~ /^E_CNJ$/) repl_pos = match(expression, /@[0-9]+\^/);
    if (expr_type ~ /^E_PAR$/) repl_pos = match(expression, /\(@[0-9]+\)/);
    if (expr_type ~ /^E_VAR$/) repl_pos = match(expression, /[a-z]+/);
    if (expr_type ~ /^E_NG1$/) repl_pos = match(expression, /^-@[0-9]+/);
    if (expr_type ~ /^E_NG2$/) repl_pos = match(expression, /[\+\-\*\/\(\[,^]-@[0-9]+/);
    if (expr_type ~ /^E_IN1$/) repl_pos = match(expression, /^[0-9]+/);
    if (expr_type ~ /^E_IN2$/) repl_pos = match(expression, /[^@0-9][0-9]+/);

    repl_pos -= 1;
    repl_len = RLENGTH;

    begin = 1;
    len = repl_pos;
    head = substr(expression, begin, len);

    begin += repl_pos;
    len = repl_len;
    replace = substr(expression, begin, len);

    begin += repl_len;
    len = expr_len - begin + 1;
    tail = substr(expression, begin, len);

    if (debug) printf("INSTR:head = <%s>\n", head);
    if (debug) printf("INSTR:repl = <%s>\n", replace);
    if (debug) printf("INSTR:tail = <%s>\n", tail);
}

function finish() {
    print "DONE:";
    fflush();
    next;
}

function postamble(slot_num) {
    print "EXPR:" (slot_num ? (head "@" slot_num tail) : "");
    finish();
}

function postamble_2(slot_num_1, op, slot_num_2) {
    print "EXPR:" head "@" slot_num_1 op "@" slot_num_2 tail;
    finish();
}

# QUIT:

$1 ~ /^quit$/ {
    printf("INSTR:quit\n");
    finish();
}

# SLOT_MAP:

$1 ~ /^slot$/ {
    slot_map();
    printf("INSTR:ignore = prompt()\n");
    finish();
}

# DUMP:

$1 ~ /^dump$/ {
    dump_state();
    printf("INSTR:ignore = prompt()\n");
    finish();
}

# PROMPT:

$1 ~ /^$/ {
    slot_gc();
    printf("INSTR:ignore = prompt()\n");
    finish();
}

# DISPLAY:

$1 ~ /^@[0-9]+$/ {
    slot = $1;
    sub(/^@/, "", slot);

    if (slot_type[slot] ~ /base/) printf("INSTR:ignore = b_display(%s)\n", slot);
    if (slot_type[slot] ~ /extn/) printf("INSTR:ignore = e_display(%s)\n", slot);

    postamble(0);
}

# PARENTHESIS:

$1 ~ /\(@[0-9]+\)/ {

    expr_split($1, "E_PAR");
    slot = replace;

    sub(/^\(@/, "", slot);
    sub(/\)$/,  "", slot);

    postamble(slot);
}

# VARIABLE:

$1 ~ /[a-z]+/ {

    expr_split($1, "E_VAR");
    var_name = replace;

    if (var_name in var_slot) {
        slot = var_slot[var_name];
    } else {
        slot = slot_alloc("var");
        var_slot[var_name] = slot;
        slot_var_name[slot] = var_name;
    }

    postamble(slot);
}

# INITIALIZE:

# The block below picks up literal values made up of digit strings
# which do *not* begin with a @.  This block handles digit strings
# that begin at the beginning of the expression (this is the first
# case) as well as digit strings that begin somewhere inside the
# expression (this is the second case).


$1 ~ /(^|[^@0-9])[0-9]+/ {
    expr_split($1, "E_IN1");

    if (replace ~ /^$/) {
       expr_split($1, "E_IN2");
       head = head substr(replace, 1, 1);
       sub(/^./, "", replace);
    }

    literal = replace;

    slot = slot_alloc("temp");
    slot_type[slot] = "base";
    printf("INSTR:ignore = b_initialize(%s,%s)\n", slot, literal);

    postamble(slot);
}

# COPY:

$1 ~ /^@[0-9]+=@[0-9]+$/ {
    if (debug) printf("INSTR:copy\n");

    split($1, part, /=/);
    sub(/^@/, "", part[1]);
    sub(/^@/, "", part[2]);
    lvar = part[1];
    rvar = part[2];

    if (debug == 1) printf("INSTR:lvar = <%s>\n", lvar);
    if (debug == 1) printf("INSTR:rvar = <%s>\n", rvar);

    if (slot_status[lvar] == "var") {
        if (slot_type[rvar] == "base") { f = "b_copy"; }
        if (slot_type[rvar] == "extn") { f = "e_copy"; }
        printf("INSTR:ignore = %s(%s,%s)\n", f, lvar, rvar);
	slot_type[lvar] = slot_type[rvar];
    } else {
        printf("INSTR:print \"error: left operand is not a variable\\n\"\n");
    }

    postamble(0);
}

# PRODUCT:

$1 ~ /@[0-9]+\*@[0-9]+/ {

    expr_split($1, "E_PRD");

    split(replace, part, /\*/);
    sub(/^@/, "", part[1]);
    sub(/^@/, "", part[2]);

    l_arg = part[1];
    r_arg = part[2];

    if (slot_type[l_arg] == "base" && slot_type[r_arg] == "base") {
        f = "b_product";
        derived_slot_type = "base";
    } else {
        f = "e_product";
        derived_slot_type = "extn";
    }

    if (slot_type[l_arg] == "base" && slot_type[r_arg] == "extn") {
        temp_slot = slot_alloc("temp");
        printf("INSTR:ignore = e_upconvert(%s,%s)\n", temp_slot, l_arg);
        l_arg = temp_slot;
    }

    if (slot_type[l_arg] == "extn" && slot_type[r_arg] == "base") {
        temp_slot = slot_alloc("temp");
        printf("INSTR:ignore = e_upconvert(%s,%s)\n", temp_slot, r_arg);
        r_arg = temp_slot;
    }

    slot = slot_alloc("temp");
    printf("INSTR:ignore = %s(%s,%s,%s)\n", f, slot, l_arg, r_arg);
    slot_type[slot] = derived_slot_type;

    postamble(slot);
}

# DIVIDE:

$1 ~ /@[0-9]+\/@[0-9]+/ {

    expr_split($1, "E_DIV");

    split(replace, part, /\//);
    sub(/^@/, "", part[1]);
    sub(/^@/, "", part[2]);

    l_arg = part[1];
    r_arg = part[2];

    if (slot_type[r_arg] == "base") { f = "b_invert"; }
    if (slot_type[r_arg] == "extn") { f = "e_invert"; }
    slot = slot_alloc("temp");
    slot_type[slot] = slot_type[r_arg];
    printf("INSTR:ignore = %s(%s,%s)\n", f, slot, r_arg);

    postamble_2(l_arg, "*", slot);
}

# NEGATE:

$1 ~ /(^|[\+\-\*\/\(\[,^])-@[0-9]+/ {

    expr_split($1, "E_NG1");

    if (replace ~ /^$/) {
       expr_split($1, "E_NG2");
       head = head substr(replace, 1, 1);
       sub(/^./, "", replace);
    }

    rvar = replace;
    sub(/^-@/, "", rvar);

    if (slot_type[rvar] == "base") { f = "b_negate"; }
    if (slot_type[rvar] == "extn") { f = "e_negate"; }
    slot = slot_alloc("temp");
    slot_type[slot] = slot_type[rvar];
    printf("INSTR:ignore = %s(%s,%s)\n", f, slot, rvar);

    postamble(slot);
}

# SUM:

$1 ~ /@[0-9]+\+@[0-9]+/ {

    expr_split($1, "E_SUM");

    split(replace, part, /\+/);
    sub(/^@/, "", part[1]);
    sub(/^@/, "", part[2]);

    l_arg = part[1];
    r_arg = part[2];

    if (slot_type[l_arg] == "base" && slot_type[r_arg] == "base") {
	f = "b_sum";
	derived_slot_type = "base";
    } else {
	f = "e_sum";
	derived_slot_type = "extn";
    }

    if (slot_type[l_arg] == "base" && slot_type[r_arg] == "extn") {
	temp_slot = slot_alloc("temp");
	printf("INSTR:ignore = e_upconvert(%s,%s)\n", temp_slot, l_arg);
	l_arg = temp_slot;
    }

    if (slot_type[l_arg] == "extn" && slot_type[r_arg] == "base") {
	temp_slot = slot_alloc("temp");
	printf("INSTR:ignore = e_upconvert(%s,%s)\n", temp_slot, r_arg);
	r_arg = temp_slot;
    }

    slot = slot_alloc("temp");
    printf("INSTR:ignore = %s(%s,%s,%s)\n", f, slot, l_arg, r_arg);
    slot_type[slot] = derived_slot_type;

    postamble(slot);
}

# DIFFERENCE:

$1 ~ /@[0-9]+-@[0-9]+/ {

    expr_split($1, "E_DIF");

    split(replace, part, /-/);
    sub(/^@/, "", part[1]);
    sub(/^@/, "", part[2]);

    l_arg = part[1];
    r_arg = part[2];

    if (slot_type[l_arg] == "base" && slot_type[r_arg] == "base") {
	f = "b_difference";
	derived_slot_type = "base";
    } else {
	f = "e_difference";
	derived_slot_type = "extn";
    }

    if (slot_type[l_arg] == "base" && slot_type[r_arg] == "extn") {
	temp_slot = slot_alloc("temp");
	printf("INSTR:ignore = e_upconvert(%s,%s)\n", temp_slot, l_arg);
	l_arg = temp_slot;
    }

    if (slot_type[l_arg] == "extn" && slot_type[r_arg] == "base") {
	temp_slot = slot_alloc("temp");
	printf("INSTR:ignore = e_upconvert(%s,%s)\n", temp_slot, r_arg);
	r_arg = temp_slot;
    }

    slot = slot_alloc("temp");
    printf("INSTR:ignore = %s(%s,%s,%s)\n", f, slot, l_arg, r_arg);
    slot_type[slot] = derived_slot_type;

    postamble(slot);
}

# EXTENSION LITERAL:

$1 ~ /\[@[0-9]+(,@[0-9]+)*\]/ {

    expr_split($1, "E_EXT");

    slot = slot_alloc("temp");
    slot_type[slot] = "extn";

    printf("INSTR:ignore = e_reset(%s)\n", slot);
    sub(/\[/, "", replace);
    sub(/\]/, "", replace);
    n = split(replace, part, /,/);
    for (i=1; i<=n; ++i) {
        sub(/^@/, "", part[i]);
        printf("INSTR:ignore = e_initialize(%s,%s,%s)\n", slot, part[i], i);
    }

    postamble(slot);
}

# CONJUGATE

$1 ~ /@[0-9]+\^/ {

    expr_split($1, "E_CNJ");

    sub(/^@/, "", replace);
    sub(/\^/, "", replace);

    arg = replace;

    slot = slot_alloc("temp");
    if (slot_type[arg] == "extn") {
        printf("INSTR:ignore = e_conjugate(%s,%s)\n", slot, arg);
        slot_type[slot] = "extn";
    } else {
        printf("INSTR:ignore = error_message(%s)\n", 0);
    }

    postamble(slot);
}

END {
    dump_state();
}

'
