#!/usr/bin/bash

debug=1

preamble=$(
    if [[ -e diag/front-end.state.awk ]]
    then
	cat diag/front-end.state.awk
    else
	echo "next_slot_loc = 10;"
    fi
)

awk "

BEGIN {
    $preamble
    debug = $debug;
}

"'function slot_alloc(ref_type) {
    new_slot = -1;

    for (new_slot in slot_free) {
        break;
    }

    if (new_slot == -1) {
        new_slot = next_slot_loc;
        ++ next_slot_loc;
    } else {
        delete slot_free[new_slot];
    }

    if (ref_type ~ /temp/) {
        slot_temp[new_slot] = 1;
        slot_status[new_slot] = "temp";
    }

    if (ref_type ~ /var/) {
        slot_var[new_slot] = 1;
        slot_status[new_slot] = "var";
    }

    return new_slot;
}

function dump_state() {
    printf("next_slot_loc=%s\n", next_slot_loc) > "diag/front-end.state.awk";
    for (s in slot_free) printf("slot_free[%s]=1\n", s) >> "diag/front-end.state.awk";
    for (s in slot_temp) printf("slot_temp[%s]=1\n", s) >> "diag/front-end.state.awk";
    for (s in slot_var ) printf("slot_var[%s]=1\n", s) >> "diag/front-end.state.awk";
    for (s in slot_status) printf("slot_status[%s]=\"%s\"\n", s, slot_status[s]) >> "diag/front-end.state.awk";
    for (s in slot_var_name) printf("slot_var_name[%s]=\"%s\"\n", s, slot_var_name[s]) >> "diag/front-end.state.awk";
    for (n in var_slot) printf("var_slot[\"%s\"]=%s\n", n, var_slot[n]) >> "diag/front-end.state.awk";
    for (s in slot_type) printf("slot_type[%s]=\"%s\"\n", s, slot_type[s]) >> "diag/front-end.state.awk";
}

function slot_map() {
    printf("INSTR:print \"next_slot_loc=\",%s,\"\\n\"\n", next_slot_loc);
    for (i=0; i<next_slot_loc; ++i) {
        if (slot_status[i] == "var") {
            printf("INSTR:print \"slot[%s]=%s:%s\\n\"\n",
                   i, slot_status[i], slot_var_name[i]);
        } else {
            printf("INSTR:print \"slot[%s]=%s\\n\"\n", i, slot_status[i]);
        }
    }
}

function slot_gc() {
    for (slot_loc in slot_temp) {
        slot_status[slot_loc] = "free";
        slot_free[slot_loc] = 1;
    }

    delete slot_temp;
}

# QUIT:

$1 ~ /^quit$/ {
    printf("INSTR:quit\n");
    print "DONE:";
    fflush();
    next;
}

# SLOT_MAP:

$1 ~ /^slot$/ {
    slot_map();
    printf("INSTR:ignore = prompt()\n");
    print "DONE:";
    fflush();
    next;
}

# DUMP:

$1 ~ /^dump$/ {
    dump_state();
    printf("INSTR:ignore = prompt()\n");
    print "DONE:";
    fflush();
    next;
}

# PROMPT:

$1 ~ /^$/ {
    slot_gc();
    printf("INSTR:ignore = prompt()\n");
    print "DONE:";
    fflush();
    next;
}

# DISPLAY:

$1 ~ /^@[0-9]+$/ {
    i = $1;
    sub(/^@/,"",i);
    if (slot_type[i] ~ /base/) printf("INSTR:ignore = b_display(%s)\n",i);
    if (slot_type[i] ~ /extn/) printf("INSTR:ignore = e_display(%s)\n",i);
    print "EXPR:";
    print "DONE:";
    fflush();
    next;
}

# PARENTHESIS:

$1 ~ /\(@[0-9]+\)/ {
    if (debug) printf("INSTR:parenthesis\n");

    expr = $1;
    expr_len = length(expr);

    replace_pos = match(expr, /\(@[0-9]+\)/);
    replace_len = RLENGTH;

    begin = 1;
    len = replace_pos - 1;
    head = substr(expr, begin, len);

    begin += len + 1;
    len = replace_len - 2;
    subexpr = substr(expr, begin, len);

    begin += replace_len - 1;
    len = expr_len - begin + 1;
    tail = substr(expr, begin, len);

    if (debug) printf("INSTR:head = <%s>\n", head);
    if (debug) printf("INSTR:subexpr = <%s>\n", subexpr);
    if (debug) printf("INSTR:tail = <%s>\n", tail);

    new_expr = head subexpr tail;
    print "EXPR:" new_expr;
    print "DONE:";
    fflush();
    next;
}

# VARIABLE:

$1 ~ /[a-z]+/ {
    if (debug) printf("INSTR:variable\n");

    expr = $1;
    expr_len = length(expr);

    replace_pos = match(expr, /[a-z]+/);
    replace_len = RLENGTH;

    begin = 1;
    len = replace_pos - 1;
    head = substr(expr, begin, len);

    begin += len;
    len = replace_len;
    var_name = substr(expr, begin, len);

    begin += replace_len;
    len = expr_len - begin + 1;
    tail = substr(expr, begin, len);

    if (debug) printf("INSTR:head = <%s>\n", head);
    if (debug) printf("INSTR:variable = <%s>\n", var_name);
    if (debug) printf("INSTR:tail = <%s>\n", tail);

    if (var_name in var_slot) {
        reference = "@" var_slot[var_name];
    } else {
        slot = slot_alloc("var");
        var_slot[var_name] = slot;
        slot_var_name[slot] = var_name;
        reference = "@" slot;
    }
    new_expr = head reference tail;
    print "EXPR:" new_expr;
    print "DONE:";
    fflush();
    next;
}

# INITIALIZE:

# The block below picks up literal values made up of digit strings
# which do *not* begin with a @.  This block handles digit strings
# that begin somewhere inside the expression (this is the first case)
# as well as digit strings that begin at the beginning of the expression
# (this is the second case).

$1 ~ /[^@0-9][0-9]+/ || $1 ~ /^[0-9]+/ {
    if (debug) printf("INSTR:initialize\n");

    expr = $1;
    expr_len = length(expr);

    if ($1 ~ /[^@0-9][0-9]+/) {
	literal_pos = match(expr, /[^@0-9][0-9]+/);
	literal_len = RLENGTH - 1;

	begin = 1;
	len = literal_pos;
	head = substr(expr, begin, len);

	begin += literal_pos;
	len = literal_len;
	literal = substr(expr, begin, len);

	begin += literal_len;
	len = expr_len - begin + 1;
	tail = substr(expr, begin, len);
    } else {
	literal_pos = match(expr, /^[0-9]+/);
	literal_len = RLENGTH;

	head = "";

	begin = 1;
	len = literal_len;
	literal = substr(expr, begin, len);

	begin += literal_len;
	len = expr_len - begin + 1;
	tail = substr(expr, begin, len);
    }

    if (debug == 1) printf("INSTR:head = <%s>\n", head);
    if (debug == 1) printf("INSTR:literal = <%s>\n", literal);
    if (debug == 1) printf("INSTR:tail = <%s>\n", tail);

    slot = slot_alloc("temp");
    slot_type[slot] = "base";
    printf("INSTR:ignore = b_initialize(%s,%s)\n", slot, literal);

    reference = "@" slot;
    new_expr = head reference tail;
    print "EXPR:" new_expr;
    print "DONE:";
    fflush();
    next;
}

# COPY:

$1 ~ /^@[0-9]+=@[0-9]+$/ {
    if (debug) printf("INSTR:copy\n");

    lvar = $1;
    sub(/=.*$/, "", lvar);
    sub(/@/, "", lvar);
    rvar = $1;
    sub(/^.*=/, "", rvar);
    sub(/@/, "", rvar);

    if (debug == 1) printf("INSTR:lvar = <%s>\n", lvar);
    if (debug == 1) printf("INSTR:rvar = <%s>\n", rvar);

    if (slot_status[lvar] == "var") {
        if (slot_type[rvar] == "base") { f = "b_copy"; }
        if (slot_type[rvar] == "extn") { f = "e_copy"; }
        printf("INSTR:ignore = %s(%s,%s)\n", f, lvar, rvar);
	slot_type[lvar] = slot_type[rvar];
    } else {
        printf("INSTR:print \"error: left operand is not a variable\\n\"\n");
    }
    print "EXPR:";
    print "DONE:";
    fflush();
    next;
}

# PRODUCT:

$1 ~ /@[0-9]+\*@[0-9]+/ {
    if (debug) printf("INSTR:product\n");

    expr = $1;
    expr_len = length(expr);

    replace_pos = match(expr, /@[0-9]+\*@[0-9]+/) - 1;
    replace_len = RLENGTH;

    begin = 1;
    len = replace_pos;
    head = substr(expr, begin, len);

    begin += replace_pos;
    len = replace_len;
    replace = substr(expr, begin, len);

    begin += replace_len;
    len = expr_len - begin + 1;
    tail = substr(expr, begin, len);

    if (debug) printf("INSTR:head = <%s>\n", head);
    if (debug) printf("INSTR:replace = <%s>\n", replace);
    if (debug) printf("INSTR:tail = <%s>\n", tail);

    split(replace, part, /\*/);
    sub(/^@/, "", part[1]);
    sub(/^@/, "", part[2]);

    l_arg = part[1];
    r_arg = part[2];

    slot = slot_alloc("temp");
    if (slot_type[l_arg] == "base" && slot_type[r_arg] == "base") {
        f = "b_product";
        derived_slot_type = "base";
    }
    if (slot_type[l_arg] == "extn" && slot_type[r_arg] == "extn") {
        f = "e_product";
        derived_slot_type = "extn";
    }
    if (slot_type[l_arg] == "base" && slot_type[r_arg] == "extn") {
	f = "e_product";
        derived_slot_type = "extn";
        temp_slot = slot_alloc("temp");
        printf("INSTR:ignore = e_upconvert(%s,%s)\n", temp_slot, l_arg);
        l_arg=temp_slot;
    }
    if (slot_type[l_arg] == "extn" && slot_type[r_arg] == "base") {
	f = "e_product";
        derived_slot_type = "extn";
        temp_slot = slot_alloc("temp");
        printf("INSTR:ignore = e_upconvert(%s,%s)\n", temp_slot, r_arg);
        r_arg=temp_slot;
    }
    printf("INSTR:ignore = %s(%s,%s,%s)\n", f, slot, l_arg, r_arg);
    slot_type[slot] = derived_slot_type;

    reference = "@" slot;
    new_expr = head reference tail;
    print "EXPR:" new_expr;
    print "DONE:";
    fflush();
    next;
}

# DIVIDE:

$1 ~ /@[0-9]+\/@[0-9]+/ {
    if (debug) printf("INSTR:divide\n");

    expr = $1;
    expr_len = length(expr);

    replace_pos = match(expr, /@[0-9]+\/@[0-9]+/) - 1;
    replace_len = RLENGTH;

    begin = 1;
    len = replace_pos;
    head = substr(expr, begin, len);

    begin += replace_pos;
    len = replace_len;
    replace = substr(expr, begin, len);

    begin += replace_len;
    len = expr_len - begin + 1;
    tail = substr(expr, begin, len);

    if (debug) printf("INSTR:head = <%s>\n", head);
    if (debug) printf("INSTR:replace = <%s>\n", replace);
    if (debug) printf("INSTR:tail = <%s>\n", tail);

    split(replace, part, /\//);
    sub(/^@/, "", part[1]);
    sub(/^@/, "", part[2]);

    l_arg = part[1];
    r_arg = part[2];

    slot = slot_alloc("temp");
    if (slot_type[l_arg] == "base" && slot_type[r_arg] == "base") {
        f = "b_divide";
        derived_slot_type = "base";
    }
    if (slot_type[l_arg] == "extn" && slot_type[r_arg] == "base") {
        f = "e_product";
        derived_slot_type = "extn";
        temp_slot_1 = slot_alloc("temp");
        temp_slot_2 = slot_alloc("temp");
        printf("INSTR:ignore = b_invert(%s,%s)\n", temp_slot_1, r_arg);
        printf("INSTR:ignore = e_upconvert(%s,%s)\n", temp_slot_2, temp_slot_1);
        l_arg=temp_slot_2;
    }
    if (slot_type[l_arg] == "base" && slot_type[r_arg] == "extn") {
        f = "e_product";
        derived_slot_type = "extn";
        temp_slot_1 = slot_alloc("temp");
        temp_slot_2 = slot_alloc("temp");
        printf("INSTR:ignore = e_upconvert(%s,%s)\n", temp_slot_1, l_arg);
        printf("INSTR:ignore = e_invert(%s,%s)\n", temp_slot_2, r_arg);
        l_arg=temp_slot_1;
        r_arg=temp_slot_2;
    }
    if (slot_type[l_arg] == "extn" && slot_type[r_arg] == "extn") {
        f = "e_product";
        derived_slot_type = "extn";
        temp_slot_1 = slot_alloc("temp");
        printf("INSTR:ignore = e_invert(%s,%s)\n", temp_slot_1, r_arg);
	r_arg = temp_slot_1;
    }
    printf("INSTR:ignore = %s(%s,%s,%s)\n", f, slot, l_arg, r_arg);
    slot_type[slot] = derived_slot_type;

    reference = "@" slot;
    new_expr = head reference tail;
    print "EXPR:" new_expr;
    print "DONE:";
    fflush();
    next;
}

# SUM:

$1 ~ /@[0-9]+\+@[0-9]+/ {
    if (debug) printf("INSTR:sum\n");

    expr = $1;
    expr_len = length(expr);

    replace_pos = match(expr, /@[0-9]+\+@[0-9]+/) - 1;
    replace_len = RLENGTH;

    begin = 1;
    len = replace_pos;
    head = substr(expr, begin, len);

    begin += replace_pos;
    len = replace_len;
    replace = substr(expr, begin, len);

    begin += replace_len;
    len = expr_len - begin + 1;
    tail = substr(expr, begin, len);

    if (debug) printf("INSTR:head = <%s>\n", head);
    if (debug) printf("INSTR:replace = <%s>\n", replace);
    if (debug) printf("INSTR:tail = <%s>\n", tail);

    split(replace, part, /\+/);
    sub(/^@/, "", part[1]);
    sub(/^@/, "", part[2]);

    l_arg = part[1];
    r_arg = part[2];

    slot = slot_alloc("temp");
    if (slot_type[l_arg] == "base" && slot_type[r_arg] == "base") {
        f = "b_sum";
        derived_slot_type = "base";
    }
    if (slot_type[l_arg] == "extn" && slot_type[r_arg] == "extn") {
        f = "e_sum";
        derived_slot_type = "extn";
    }
    if (slot_type[l_arg] == "base" && slot_type[r_arg] == "extn") {
	f = "e_sum";
        derived_slot_type = "extn";
        temp_slot = slot_alloc("temp");
        printf("INSTR:ignore = e_upconvert(%s,%s)\n", temp_slot, l_arg);
        l_arg=temp_slot;
    }
    if (slot_type[l_arg] == "extn" && slot_type[r_arg] == "base") {
	f = "e_sum";
        derived_slot_type = "extn";
        temp_slot = slot_alloc("temp");
        printf("INSTR:ignore = e_upconvert(%s,%s)\n", temp_slot, r_arg);
        r_arg=temp_slot;
    }
    printf("INSTR:ignore = %s(%s,%s,%s)\n", f, slot, l_arg, r_arg);
    slot_type[slot] = derived_slot_type;

    reference = "@" slot;
    new_expr = head reference tail;
    print "EXPR:" new_expr;
    print "DONE:";
    fflush();
    next;
}

# NEGATE:

$1 ~ /^-@[0-9]+/ || $1 ~ /[^0-9]-@[0-9]+/ {
    if (debug) printf("INSTR:negate\n");

    expr = $1;
    expr_len = length(expr);

    replace_pos = match(expr, /^-@[0-9]+/);
    if (replace_pos >= 1) {
        replace_pos -= 1;
        replace_len = RLENGTH;
    } else {
        replace_pos = match(expr, /[^0-9]-@[0-9]+/);
        replace_len = RLENGTH - 1;
    }

    begin = 1;
    len = replace_pos;
    head = substr(expr, begin, len);

    begin += replace_pos;
    len = replace_len;
    replace = substr(expr, begin, len);

    begin += replace_len;
    len = expr_len - begin + 1;
    tail = substr(expr, begin, len);

    if (debug == 1) printf("INSTR:head = <%s>\n", head);
    if (debug == 1) printf("INSTR:replace = <%s>\n", replace);
    if (debug == 1) printf("INSTR:tail = <%s>\n", tail);

    rvar = substr(replace,3,replace_len-2);

    slot = slot_alloc("temp");
    if (slot_type[rvar] == "base") { f = "b_negate"; }
    if (slot_type[rvar] == "extn") { f = "e_negate"; }
    printf("INSTR:ignore = %s(%s,%s)\n", f, slot, rvar);
    slot_type[slot] = slot_type[rvar];

    reference = "@" slot;
    new_expr = head reference tail;
    print "EXPR:" new_expr;
    print "DONE:";
    fflush();
    next;
}

# DIFFERENCE:

$1 ~ /@[0-9]+-@[0-9]+/ {
    if (debug) printf("INSTR:difference\n");

    expr = $1;
    expr_len = length(expr);

    replace_pos = match(expr, /@[0-9]+-@[0-9]+/) - 1;
    replace_len = RLENGTH;

    begin = 1;
    len = replace_pos;
    head = substr(expr, begin, len);

    begin += replace_pos;
    len = replace_len;
    replace = substr(expr, begin, len);

    begin += replace_len;
    len = expr_len - begin + 1;
    tail = substr(expr, begin, len);

    if (debug) printf("INSTR:head = <%s>\n", head);
    if (debug) printf("INSTR:replace = <%s>\n", replace);
    if (debug) printf("INSTR:tail = <%s>\n", tail);

    split(replace, part, /-/);
    sub(/^@/, "", part[1]);
    sub(/^@/, "", part[2]);

    l_arg = part[1];
    r_arg = part[2];

    slot = slot_alloc("temp");
    if (slot_type[l_arg] == "base" && slot_type[r_arg] == "base") {
        f = "b_difference";
        derived_slot_type = "base";
    }
    if (slot_type[l_arg] == "extn" && slot_type[r_arg] == "extn") {
        f = "e_difference";
        derived_slot_type = "extn";
    }
    if (slot_type[l_arg] == "base" && slot_type[r_arg] == "extn") {
	f = "e_difference";
        derived_slot_type = "extn";
        temp_slot = slot_alloc("temp");
        printf("INSTR:ignore = e_upconvert(%s,%s)\n", temp_slot, l_arg);
        l_arg=temp_slot;
    }
    if (slot_type[l_arg] == "extn" && slot_type[r_arg] == "base") {
	f = "e_difference";
        derived_slot_type = "extn";
        temp_slot = slot_alloc("temp");
        printf("INSTR:ignore = e_upconvert(%s,%s)\n", temp_slot, r_arg);
        r_arg=temp_slot;
    }
    printf("INSTR:ignore = %s(%s,%s,%s)\n", f, slot, l_arg, r_arg);
    slot_type[slot] = derived_slot_type;

    reference = "@" slot;
    new_expr = head reference tail;
    print "EXPR:" new_expr;
    print "DONE:";
    fflush();
    next;
}

# EXTENSION LITERAL:

$1 ~ /\[@[0-9]+(,@[0-9]+)*\]/ {
    if (debug) printf("INSTR:extension literal\n");

    expr = $1;
    expr_len = length(expr);

    replace_pos = match(expr, /\[@[0-9]+(,@[0-9]+)*\]/) - 1;
    replace_len = RLENGTH;

    begin = 1;
    len = replace_pos;
    head = substr(expr, begin, len);

    begin += replace_pos;
    len = replace_len;
    replace = substr(expr, begin, len);

    begin += replace_len;
    len = expr_len - begin + 1;
    tail = substr(expr, begin, len);

    if (debug) printf("INSTR:head = <%s>\n", head);
    if (debug) printf("INSTR:replace = <%s>\n", replace);
    if (debug) printf("INSTR:tail = <%s>\n", tail);

    slot = slot_alloc("temp");
    slot_type[slot] = "extn";

    printf("INSTR:ignore = e_reset(%s)\n", slot);
    sub(/\[/, "", replace);
    sub(/\]/, "", replace);
    n = split(replace, part, /,/);
    for (i=1; i<=n; ++i) {
        sub(/^@/, "", part[i]);
        printf("INSTR:ignore = e_initialize(%s,%s,%s)\n", slot, part[i], i);
    }

    reference = "@" slot;
    new_expr = head reference tail;
    print "EXPR:" new_expr;
    print "DONE:";
    fflush();
    next;
}

# CONJUGATE

$1 ~ /@[0-9]+\^/ {
    if (debug) printf("INSTR:conjugate\n");

    expr = $1;
    expr_len = length(expr);

    replace_pos = match(expr, /@[0-9]+\^/) - 1;
    replace_len = RLENGTH;

    begin = 1;
    len = replace_pos;
    head = substr(expr, begin, len);

    begin += replace_pos;
    len = replace_len;
    replace = substr(expr, begin, len);

    begin += replace_len;
    len = expr_len - begin + 1;
    tail = substr(expr, begin, len);

    if (debug) printf("INSTR:head = <%s>\n", head);
    if (debug) printf("INSTR:replace = <%s>\n", replace);
    if (debug) printf("INSTR:tail = <%s>\n", tail);

    sub(/^@/, "", replace);
    sub(/\^/, "", replace);

    arg = replace;

    slot = slot_alloc("temp");
    if (slot_type[arg] == "extn") {
        printf("INSTR:ignore = e_conjugate(%s,%s)\n", slot, arg);
        slot_type[slot] = "extn";
    } else {
        printf("INSTR:ignore = error_message(%s)\n", 0);
    }

    reference = "@" slot;
    new_expr = head reference tail;
    print "EXPR:" new_expr;
    print "DONE:";
    fflush();
    next;
}

END {
    dump_state();
}

'
